/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/ 
/*
  Copyright (C) 1998-1999 Johan Redestig
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
****************************************************************************/
#include <algorithm>
#include "stdafx.h"
#include "PopupMaker.h"
#include "PopupMenu.h"
#include "DirectoryFolder.h"
#include "Popup.h"
PopupMaker* pPopupMaker;

void PopupBangCommand(HWND hCaller, const char *pszName, const char *pszArguments);

int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	g_hResStrInstance = dllInst;
	CoInitialize(NULL);
	pPopupMaker = new PopupMaker();
	pPopupMaker->Initialize(ParentWnd, dllInst, szPath);

	AddBangCommandEx(BANG(""), PopupBangCommand);
    AddBangCommand(BANG("DynamicFolder"), DynamicBangCommand);
	return 0;
}

int quitModule(HINSTANCE dllInst)
{
	RemoveBangCommand(BANG("DynamicFolder"));
    RemoveBangCommand(BANG(""));
	CoUninitialize();
	delete pPopupMaker;

	return 1;
}

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall DllMain(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// I believe this pretty much what the default DllMain does
	// see (MSVCDIR)\crt\src\dllmain.c for the default
	if (DLL_PROCESS_ATTACH  == fdwReason)
		DisableThreadLibraryCalls(hInst);
   	return TRUE;
}

// Dummy startup routine that does almost nothing.
// This cuts out all of the standard startup code crud that
// bloats the DLL, and makes it take longer to load
//BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
//{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
//	DisableThreadLibraryCalls(hInst);
//	return TRUE;
//}

void ForceSetForegroundWindow(HWND hWnd)
{
   DWORD currentForeground = GetWindowThreadProcessId(GetForegroundWindow(), NULL);
   DWORD popupThread = GetCurrentThreadId();

   if (currentForeground != popupThread)
   {
      // If we're not the foreground, join it
      AttachThreadInput(currentForeground, popupThread, TRUE);
   }
   
   // Either way we need to bring our window to the front
   SetForegroundWindow(hWnd);
   SetFocus(hWnd); // Weird. Without this it only works the first time.

   if (currentForeground != popupThread)
   {
      // Detach the threads and put everything back like we found it.
      AttachThreadInput(currentForeground, popupThread, FALSE);
   }

   // Redraw the popup. Not sure if we need this, but doesn't hurt.
   InvalidateRect(hWnd, NULL, TRUE);
}

template <typename data_t>
class MaybeParam
{
  bool bHaveIt_;
  data_t xValue_;
public:
  MaybeParam(void) : bHaveIt_(false) { }
  MaybeParam(const data_t &xDefault) : bHaveIt_(false), xValue_(xDefault) { }

  const MaybeParam &operator=(const data_t &xValue)
  { xValue_ = xValue; bHaveIt_ = true; return *this; }

  bool HaveIt(void) const { return bHaveIt_; }
  const data_t &Value(void) const { return xValue_; }
  operator const data_t &() const { return Value(); }
};

/* parse parameters and show menu */ 
void DoBangPopup(LPCSTR sArguments, PopupMenu *pMenu)
{
  /* it's not really safe to use a buffer <4096 bytes with the tokenizers */ 
  /* LCTokenize would require all of them to exist at the same time, so   */ 
  /* to be safe, GetToken will require less stack space                   */ 
  char token[MAX_LINE_LENGTH];
  LPCSTR sNextArg = sArguments;

  MaybeParam<int> x, y, xAlign(ALIGN_LEFT), yAlign(ALIGN_TOP);
  bool bClose       = false;
  bool bPin         = false;

  /* parse arguments */ 
  while (GetToken(sNextArg, token, &sNextArg, FALSE))
  {
    if (isdigit(*token) || *token=='-')
    {
      /* coordinates on other monitors would be complicated to get right */ 
      if (!x.HaveIt())
        x = ParseCoordinate(token, 0, GetSystemMetrics(SM_CXSCREEN));
      else if (!y.HaveIt())
        y = ParseCoordinate(token, 0, GetSystemMetrics(SM_CYSCREEN));
    }
    else if (strnicmp("x=", token, 2)==0)
      x = ParseCoordinate(token+2, 0, GetSystemMetrics(SM_CXSCREEN));
    else if (strnicmp("y=", token, 2)==0)
      y = ParseCoordinate(token+2, 0, GetSystemMetrics(SM_CYSCREEN));
    else if (stricmp("pin", token)==0)
      bPin = true;
    else if (stricmp("close", token)==0)
      bClose = true;
    else if (stricmp("left", token)==0)
      xAlign = ALIGN_LEFT;
    else if (stricmp("right", token)==0)
      xAlign = ALIGN_RIGHT;
    else if (stricmp("top", token)==0)
      yAlign = ALIGN_TOP;
    else if (stricmp("bottom", token)==0)
      yAlign = ALIGN_BOTTOM;
    else if (stricmp("center", token)==0)
    {
      /* center fills in whatever needs to be filled in */ 
      /* this way order doesn't matter and a single     */ 
      /* center works in both directions.               */ 
      if (!xAlign.HaveIt()) xAlign = ALIGN_CENTER;
      if (!yAlign.HaveIt()) yAlign = ALIGN_CENTER;
    }
  }

  /* get position from mouse if not given */ 
  if (!x.HaveIt() || !y.HaveIt())
  {
    POINT Mouse;
    GetCursorPos(&Mouse);
    x = (int)Mouse.x;
    y = (int)Mouse.y;

    /* set to autoalign if nothing specific requested */ 
    if (!xAlign.HaveIt())
      xAlign = ALIGN_AUTO;
    if (!yAlign.HaveIt())
      yAlign = ALIGN_AUTO;

    /* don't popup directly under the mouse */ 
    /* (autoalign is handled later)         */ 
    switch (xAlign.Value())
    {
    case ALIGN_LEFT: x = (int)Mouse.x + 1; break;
    case ALIGN_RIGHT: x = (int)Mouse.x - 1; break;
    case ALIGN_CENTER:
      switch (yAlign.Value())
      {
      case ALIGN_TOP: y = (int)Mouse.y + 1; break;
      case ALIGN_BOTTOM: y = (int)Mouse.y - 1; break;
      }
      break;
    }
  }

  /* show or hide the menu */ 
  if (bClose)
  {
    pMenu->SetPinned(false);
    pMenu->Hide(1);
  }
  else
  {
    if (bPin)
      pMenu->SetPinned(true);
    pMenu->Show(x, y, xAlign, yAlign);
    ForceSetForegroundWindow(pMenu->GetWindow());
  }
}

void DynamicBangCommand(HWND hCaller, const char *pszArguments)
{
    char pszLocation[MAX_BANGARGS];
    LPCSTR pszNextToken;
    
    static PopupMenu* pDynMenu = NULL;
    
    if (pDynMenu)
    {
        vector<PopupMenu*>::iterator iter =
            find(pPopupMaker->m_RootMenus.begin(), pPopupMaker->m_RootMenus.end(), pDynMenu);
        
        if (iter != pPopupMaker->m_RootMenus.end())
            pPopupMaker->m_RootMenus.erase(iter);

        delete pDynMenu;
    }

    if (GetToken(pszArguments, pszLocation, &pszNextToken, false))
    {
        pDynMenu = pPopupMaker->LoadFolder(PathFindFileName(pszLocation), pszLocation);
        pPopupMaker->m_RootMenus.push_back(pDynMenu);
        DoBangPopup(pszNextToken, pDynMenu);
    }
}

// handle !PopupXXX bang commands
void PopupBangCommand(HWND hCaller, const char *pszName, const char *pszArguments)
{
  PopupMenu *pMenu = NULL;
  unsigned int i;

  /* should use stl iterators, but I'm not 100% clear on those */ 
  for (i=0; i<pPopupMaker->m_RootMenus.size(); ++i)
  {
    pMenu = pPopupMaker->m_RootMenus[i];
    const char *sName = pMenu->GetBangName();

    if (sName && stricmp(sName, pszName)==0)
      break;
  }

  if (i==pPopupMaker->m_RootMenus.size())
  {
    /* do we really care at this point? */ 
    if (stricmp("!popup", pszName)==0)
      pMenu = pPopupMaker->m_pRoot;
    else
      pMenu = NULL;
  }

  if (pMenu)
    DoBangPopup(pszArguments, pMenu);
}
